/**
 * @brief This example demonstrates advanced processing of the incoming SMS message in the 
 * Adeon format. The target is to set the values of defined parameters via SMS in case that
 * the SMS has been sent by authorized user (from authorized phone number).
 * 
 *  Copyright (c) 2019 JSC electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <Arduino.h>
#include <SoftwareSerial.h>
#include <AdeonGSM.h>
#include <utility/SIMlib.h>

#define getName(var)  #var 

#define INPUT_SIGNAL_OPEN   20
#define INPUT_SIGNAL_CLOSE  21

#define OUTPUT_SIGNAL       6

#define CHECK_PERIOD        60

#define RX                  10
#define TX                  11

SoftwareSerial gsmSerial = SoftwareSerial(RX, TX);
GSM* gsm;
Adeon adeon;

typedef enum{
    UNKNOWN = 0,
    OPEN,
    CLOSE
}GateState;

uint16_t counterAutoClose = 0;
uint16_t counterCheckState = 0;

const char* pnHost = "420111111111";
const char* pnUser = "420222222222";
const char* pnAdmin = "420333333333";

const char* parGate = "GATE";
const char* parClose = "CLOSE";
const char* parAccess = "ACCESS";

char* msgBuf; 
char* pnBuf; 

GateState gatePos;

void callbackGateImpulse(uint16_t val);
void gateISR();
void accessManagement();
void setupTimer();
void userInit();
void paramInit();
void numOfItems();
void processMsg();

void callbackGateImpulse(uint16_t val){
    //callback function which is called if value of parameter is edited
    Serial.print(F("OUTPUT VAL: "));
    Serial.println(val);

    if(val > 0){
        digitalWrite(OUTPUT_SIGNAL, HIGH);
        delay(200);
        Serial.println(F("GATE IMPULSE"));
        digitalWrite(OUTPUT_SIGNAL, LOW);
        gatePos = UNKNOWN;
    }
}

//pokud do 60 sekund od akce nebude brána otevrena nebo zavrena – vysli impulz
void gateISR(){
    if(digitalRead(INPUT_SIGNAL_OPEN) == HIGH){
        gatePos = OPEN;
        Serial.println(F("GATE IS OPEN"));
    }
    else if(digitalRead(INPUT_SIGNAL_CLOSE) == HIGH){
        gatePos = CLOSE;
        Serial.println(F("GATE IS CLOSED"));
    }
    else{
        gatePos = UNKNOWN;
        Serial.println(F("UNKNOWN GATE POSITION"));
    }
}

void accessManagement(){
    Serial.print(parGate);
    Serial.print(" SET TO ");
    uint16_t val = adeon.getParamValue(parAccess);
    if(val == 0){
        adeon.setParamAccess(parGate, ADEON_ADMIN);
        Serial.println(getName(ADEON_ADMIN));
    }
    else if(val == 1){
        adeon.setParamAccess(parGate, ADEON_USER);
        Serial.println(getName(ADEON_USER));
    }
    else{
        adeon.setParamAccess(parGate, ADEON_HOST);
        Serial.println(getName(ADEON_HOST));
    }
}

void setupTimer1() {
  noInterrupts();
  // Clear registers
  TCCR1A = 0;
  TCCR1B = 0;
  TCNT1 = 0;

  // 1 Hz (16000000/((15624+1)*1024))
  OCR1A = 15624;
  // CTC
  TCCR1B |= (1 << WGM12);
  // Prescaler 1024
  TCCR1B |= (1 << CS12) | (1 << CS10);
  // Output Compare Match A Interrupt Enable
  TIMSK1 |= (1 << OCIE1A);
  interrupts();
}

ISR(TIMER1_COMPA_vect) {
    if(gatePos == OPEN){
        if(adeon.getParamValue(parClose) > counterAutoClose){
            counterAutoClose++;
        }
        else{
            callbackGateImpulse(1);
        }
    }
    else{
        counterAutoClose = 0;
    }
    
    if(gatePos == UNKNOWN){
        if(CHECK_PERIOD > counterCheckState){
            counterCheckState++;
        }
        else{
            callbackGateImpulse(1);
            counterCheckState = 0;
        }
    }
    else{
        counterCheckState = 0;
    }
}

void userInit(){
    //add users with ADMIN, USER or HOST rights
    adeon.addUser(pnAdmin, ADEON_ADMIN);
    adeon.addUser(pnUser, ADEON_USER);
    adeon.addUser(pnHost, ADEON_HOST);
    adeon.printUsers();
}

void paramInit(){
    //add parameters
    adeon.addParamWithCallback(callbackGateImpulse, parGate, 0);
    adeon.addParam(parClose, 5);
    adeon.addParamWithCallback(accessManagement, parAccess, 0);
    adeon.printParams();
}

void numOfItems(){
  Serial.print(F("NUM OF USERS: "));
  Serial.println(adeon.getNumOfUsers()); //number of saved users

  Serial.print(F("NUM OF PARAMS: "));
  Serial.println(adeon.getNumOfParams()); //number of saved parameters
  Serial.println();
}

void processMsg(){
    Serial.println("\nPROCESSING INCOMMING MSG.");
    //simulation of incomming message 
    if(adeon.isAdeonReady()){
        if(adeon.isUserInAdeon(pnBuf)){
            Serial.println(F("PHONE NUMBER IS AUTHORIZED"));
            Serial.println(msgBuf);
            //parameters are parsed and their values are saved into list
            adeon.parseBuf(msgBuf, adeon.getUserRightsLevel(pnBuf));
            adeon.printParams();
        }
        else {
            Serial.println(F("PHONE NUMBER IS NOT AUTHORIZED\n"));
        }
    }
    else{
        Serial.println(F("ADEON IS NOT READY TO PROCCESS NEW MESSAGE."));
    }
}

void setup() {
    // Setup the Serial port. See http://arduino.cc/en/Serial/IfSerial
    Serial.begin(9600);
    gsmSerial.begin(9600);
    pinMode(LED_BUILTIN, OUTPUT);
    pinMode(OUTPUT_SIGNAL, OUTPUT);
    pinMode(INPUT_SIGNAL_OPEN, INPUT);
    pinMode(INPUT_SIGNAL_CLOSE, INPUT);
      
    digitalWrite(LED_BUILTIN, LOW);
    digitalWrite(OUTPUT_SIGNAL, LOW);
    delay(200);

    gsm = new GSM(&gsmSerial);
    gsm->begin();

    userInit();
    paramInit();
    numOfItems();  
    gateISR();
    setupTimer1();

    attachInterrupt(digitalPinToInterrupt(INPUT_SIGNAL_CLOSE), gateISR, CHANGE);
    attachInterrupt(digitalPinToInterrupt(INPUT_SIGNAL_OPEN), gateISR, CHANGE);  
}

void loop() {
    gsm->checkGsmOutput();
    if(gsm->isNewMsgAvailable()){
        pnBuf = gsm->getPhoneNum();
        msgBuf = gsm->getMsg();
        processMsg();
    }
}
